#!/usr/bin/env fades

import argparse
import datetime
import logging
import os
import sys
import time

import exifread  # fades
import httplib2  # fades
import humanize  # fades
import yaml  # fades
from apiclient import discovery, http  # fades google-api-python-client-py3
from oauth2client import client, tools  # fades
from oauth2client.file import Storage

# hack to fix 'apiclient' for modern Python
from collections.abc import Callable
import collections
collections.Callable = Callable

SCOPE = 'https://www.googleapis.com/auth/drive'
APPLICATION_NAME = 'Auto photo downloader'

FOLDER_MIMETYPE = 'application/vnd.google-apps.folder'
IMAGE_MIMETYPE = 'image/jpeg'

RETRY_ATTEMPTS = 3
MAX_FOLDERS = 20

# set up logging
logger = logging.getLogger("msg")
h = logging.StreamHandler()
h.setFormatter(logging.Formatter("%(asctime)s %(levelname)-12s %(message)s"))
logger.addHandler(h)
logger.setLevel(logging.WARNING)


def get_credentials(auth_flags, config):
    """Gets valid user credentials from storage.

    If nothing has been stored, or if the stored credentials are invalid,
    the OAuth2 flow is completed to obtain the new credentials.

    Returns:
        Credentials, the obtained credential.
    """
    store = Storage(config['credentials_file'])
    credentials = store.get()
    if not credentials or credentials.invalid:
        flow = client.flow_from_clientsecrets(config['client_secret_file'], SCOPE)
        flow.user_agent = APPLICATION_NAME
        credentials = tools.run_flow(flow, store, auth_flags)
        logger.info('Storing credentials to %s', config['credentials_file'])
    return credentials


def retryable(orig_func):
    def new_func(*args, **kwargs):
        for attempt in range(1, RETRY_ATTEMPTS + 1):
            try:
                result = orig_func(*args, **kwargs)
            except Exception as exc:
                logger.error("Explorer failure in %s (attempt %d): %r", orig_func, attempt, exc)
                if attempt == RETRY_ATTEMPTS:
                    raise
                time.sleep(5)
            else:
                return result
    return new_func


class Explorer:
    def __init__(self, auth_flags, config):
        credentials = get_credentials(auth_flags, config)
        http = credentials.authorize(httplib2.Http())
        self.service = discovery.build('drive', 'v3', http=http)

    @retryable
    def list_folder(self, folder_id):
        """Get the active items in a folder."""
        files = self.service.files()
        request = files.list(
            pageSize=50, q="trashed = false and '{}' in parents".format(folder_id),
            fields="nextPageToken, files(id, name, mimeType, size)")

        all_items = []
        while request is not None:
            results = request.execute()
            all_items.extend(results['files'])
            request = files.list_next(request, results)

        return all_items

    @retryable
    def download(self, file_id, dest_path):
        """Download a file."""
        fh = open(dest_path, "wb")
        request = self.service.files().get_media(fileId=file_id)
        downloader = http.MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()
            logger.debug("Downloading %r: %.2f", file_id, status.progress())
        fh.close()

    @retryable
    def delete(self, file_id):
        """Remove a file."""
        request = self.service.files().delete(fileId=file_id)
        request.execute()


def get_image_tstamp(temp_path):
    """Open an image and build a prefix using internal timestamp."""
    with open(temp_path, "rb") as fh:
        tags = exifread.process_file(fh)
    if tags:
        try:
            exif_dt = tags['EXIF DateTimeOriginal'].values
        except KeyError:
            logger.warning("EXIF entry not found for %r (%s)", temp_path, tags.keys())
            tstamp = "unknown-datetime"
        else:
            dt = datetime.datetime.strptime(exif_dt, "%Y:%m:%d %H:%M:%S")
            tstamp = dt.strftime("%Y%m%d-%H%M%S")
    else:
        logger.warning("Tags not found for %r", temp_path)
        tstamp = "unknown-datetime"
    return tstamp


def main(explorer, config_file):
    """Main entry point."""
    for drive_info in config['drives_info']:
        base_gdrive_folder = drive_info['gdrive_folder']
        base_dest_dir = drive_info['destination']
        prefix = drive_info['photo_prefix']

        logger.info("Exploring %s, downloading to %s", base_gdrive_folder, base_dest_dir)
        folders = explorer.list_folder(base_gdrive_folder)
        if len(folders) > MAX_FOLDERS:
            logger.warning(
                "Partial work, deferring %d folders for later", len(folders) - MAX_FOLDERS)
            folders = folders[:MAX_FOLDERS]

        for folder in folders:
            folder_id = folder['id']
            logger.info("Found folder %r (%s)", folder['name'], folder_id)
            assert folder['mimeType'] == FOLDER_MIMETYPE

            nodes = explorer.list_folder(folder_id)
            for node in nodes:
                node_id = node['id']
                size = int(node['size'])
                if size == 0:
                    logger.warning("Discarding empty node %r (%s)", node['name'], node_id)
                    explorer.delete(node_id)
                    continue

                humansize = humanize.naturalsize(size, binary=True)
                logger.info("Found node %r (%s) %s", node['name'], node_id, humansize)

                # download the file to a temporary name
                temp_dest_path = os.path.join(base_dest_dir, "tmp-{}.jpeg".format(node_id))
                logger.debug("Downloading node to", temp_dest_path)
                explorer.download(node_id, temp_dest_path)

                # get the final filename from the node
                if node['mimeType'] == IMAGE_MIMETYPE:
                    tstamp = get_image_tstamp(temp_dest_path)
                    final_path = os.path.join(
                        base_dest_dir, "{}-{}-{}.jpeg".format(prefix, tstamp, node_id[:5]))
                else:
                    tstamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
                    ext = node['name'].split(".")[-1]
                    final_name = "{}-{}-{}.{}".format(prefix, tstamp, node_id[:5], ext)
                    final_path = os.path.join(base_dest_dir, final_name)
                logger.debug("Renaming node to", final_path)
                os.rename(temp_dest_path, final_path)

                # remove the node in gdrive
                explorer.delete(node_id)

            # ensure folder is empty and remove
            assert explorer.list_folder(folder_id) == []
            logger.info("Deleting folder %s", folder_id)
            explorer.delete(folder_id)

    logger.info("Done.")


if __name__ == '__main__':
    # manually handling sys.argv as then it's automatically processed by Explorer
    if '-v' in sys.argv:
        sys.argv.remove('-v')
        logger.setLevel(logging.INFO)

    config_file = sys.argv.pop(1)
    with open(config_file, "rt", encoding="utf8") as fh:
        config = yaml.safe_load(fh)

    auth_flags = argparse.ArgumentParser(parents=[tools.argparser]).parse_args()
    explorer = Explorer(auth_flags, config)
    main(explorer, config_file)
